<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>JupyterLab</title>
  <meta name="viewport" content="width=device-width, initial-scale=1">
    <script id="jupyter-config-data" type="application/json">{
    {% for key, value in page_config.items() -%}
    "{{ key }}": "{{ value }}",
    {% endfor -%}
    "baseUrl": "{{ base_url }}",
    "wsUrl": "{{ ws_url }}"
  }</script>

  {% block favicon %}
  <link rel="icon" type="image/x-icon" href="{{ base_url }}static/base/images/favicon.ico" class="idle favicon">
  <link rel="" type="image/x-icon" href="{{ base_url }}static/base/images/favicon-busy-1.ico" class="busy favicon">
  {% endblock %}

</head>
<body>

<script type="text/javascript">
  /* Remove token from URL. */
  (function () {
    var location = window.location;
    var search = location.search;

    // If there is no query string, bail.
    if (search.length <= 1) {
      return;
    }

    // Rebuild the query string without the `token`.
    var query = '?' + search.slice(1).split('&')
      .filter(function (param) { return param.split('=')[0] !== 'token'; })
      .join('&');

    // Rebuild the URL with the new query string.
    var url = location.origin + location.pathname +
      (query !== '?' ? query : '') + location.hash;

    window.history.replaceState({ }, '', url);
  })();
</script>

<script type="text/javascript" src="{{base_url}}lab/static/vendors~main.b2dc432665d8c9c7cb85.js"></script><script type="text/javascript" src="{{base_url}}lab/static/main.5639f2282655ad05e575.js"></script></body>
</html>
